package winFunc;

#
# winFunc.pm
#
# [ldasm] Callback-functions for all windows
#

use strict;
use Tk;
use Tk::Toplevel;
use Tk::Text;
use Tk::FileSelect;
use Tk::DialogBox;

use winMain;
use winDlg;
use asmps;
use misc;
use config;

#-------
# Export

use Exporter;
use vars qw(@ISA @EXPORT $numRows);
@ISA = qw(Exporter);
@EXPORT = qw(&FuncImport &FuncExport &FuncInternal &StringRefs &HelpMe &About
             &HelpGotoAnchor
             &MoveRowSelection
             &OpenFile &SavePrj &OpenPrj
             &SearchResultGoto &SearchText &SearchResultWin &SearchNext
             &GotoCodeLoc &GotoCodePoint
             &JumpToReference &ReferenceCopy &CopySelCode
             &ExecuteRefOp &ExecuteRefRet);

#-----------------
# Global variables

   $numRows = 0;
my $execFilename;
my $curSelTag = '';
my %refDest = undef;
my %refDlgEntry;

#--------------------------------------------
# FuncImport(), FuncExport(), FuncInternal(),
# StringRefs(), HelpMe(), About()
#
# Shows misc. dialogs

sub FuncImport()   { $FuncImport{top}->deiconify();   }
sub FuncExport()   { $FuncExport{top}->deiconify();   }
sub FuncInternal() { $FuncInternal{top}->deiconify(); }
sub StringRefs()   { $StringRefs{top}->deiconify();   }
sub HelpMe()       { $HelpMe{top}->deiconify();       }
sub About()        { $About{dlg}->Show();             }

#-----------
# HelpJump()
#
# Called when a link is pressed in the help

sub HelpGotoAnchor()
{
   (my $cursorIdx = $HelpMe{txt}->index('current')) =~ s/\..+/\.0/;
   my $anch = $HelpMe{anchor}->{$cursorIdx};
   if ($anch eq undef) {
      print "Help-error: Missing anchor '$cursorIdx'\n";
   } else {
      my $link = $HelpMe{link}->{$anch};
      if ($link eq undef) {
         print "Help-error: Missing link '$anch' ($cursorIdx)\n";
      } else {
         $HelpMe{txt}->see($link);
      }
   }
}

#--------------------
# MoveRowSelection($)
#
# in:  special ('+1'|'-1'|'-p'|'+p') or row
# out: -

sub MoveRowSelection($)
{
   my $drow = shift;

#   print "MoveRow:\n";
#   print "NumRows:$numRows, drow:$drow\n";

   # Any row or current
   return if ($numRows == 0);

   # Check if special
   if ($drow eq '+1') {
      $drow = $winMain::currentRow+1;
   } elsif ($drow eq '-1') {
      $drow = $winMain::currentRow-1;
   } elsif ($drow eq '+p') {
      $drow = $winMain::currentRow+$config::RowsPerPage;
   } elsif ($drow eq '-p') {
      $drow = $winMain::currentRow-$config::RowsPerPage;

   # No special
   } else {
      $drow =~ s/\..+//; # remove column
   }

   # Valid row?
   if ($drow > $numRows) {
      $drow = $numRows;
   } elsif ($drow <= 0) {
      $drow = 1;
   }

   # Remove old selection
   $winMain::text->tagRemove($curSelTag,
         "$winMain::currentRow.0",($winMain::currentRow+1).".0")
      if ($curSelTag ne undef);

   # Reference ?
   my @lcnt = split(/\s+/,$winMain::text->get("$drow.1","$drow.end"));
   $refDest{'to'} = $asmps::refsType{'ref'}->{$lcnt[0]};
   if ($refDest{'to'} ne undef) {
      $curSelTag = 'overRef';
      $refDest{'type'} = ($lcnt[2] eq 'call') ? 'call' : 'jump';
   } else {
      $curSelTag = 'overNormal';
   }

   # Mark new row
   $winMain::text->tagAdd($curSelTag,"$drow.0",($drow+1).".0");
   $winMain::currentRow = $drow;
   $winMain::text->see("$drow.0");
}

#------------------
# _CleanUpWidgets()
#
# Removes the content from all widgets (text, etc.), enables writting and
# shows the progress window.
#
# in+out: -

sub _CleanUpWidgets()
{
   # Save file?
   if ($execFilename ne undef) {
      SavePrj() if (MsgBox('Save current file ?'));
   }

   # References
   for (\%FuncImport,\%FuncExport,\%FuncInternal,\%StringRefs) {
      $_->{top}->withdraw();
      $_->{listbox}->delete(0,'end');
   }

   # Asm-window
   $winMain::text->configure(-state => 'normal');
   $winMain::text->delete('1.0','end');

   # Progress-window
   $Progress{text}->delete('1.0','end');
   $Progress{top}->deiconify();
}

#---------------------
# _ProgressFinished($)
#
# Hides the progress window,disabled write-access to the asm-window and
# sets the cursor on the first row.
#
# in:  filename
# out: filename w/o path

sub _ProgressFinished($)
{
   my $fname = shift;

   # Update main-window-title
   $fname =~ s|.+/([^/]+)$|$1|;
   $winMain::win->title("$config::WinTitle [$fname]");
   $winMain::win->iconname("$config::IconName-$fname");

   # Widgets
   $winMain::text->configure(-state => 'disabled');
   $Progress{top}->withdraw();

   # Misc vars
   %refDlgEntry = undef;
   %refDest = undef;
   $curSelTag = '';
   ($numRows = $winMain::text->index('end')) =~ s/\..+//;
   $numRows--;

   # Selection to first row
   MoveRowSelection(1);

   return $fname;
}

#-----------
# OpenFile()
#
# Creates and shows the main-window

sub OpenFile()
{
   # Show dialog and abort if no file choosen
   my $ofdlg = $winMain::win->FileSelect(
      -title =>  'Select File to Disassemble',
   );
   my $efname = $ofdlg->Show();
   return if ($efname eq undef);

   # Create listing
   _CleanUpWidgets();
   CreateListing($efname);
   $execFilename = _ProgressFinished($efname);
}

#----------
# SavePrj()
#
# Saves a disassembled file

sub SavePrj()
{
   # A file available?
   return if ($execFilename eq undef);

   # Dialog
   my $sfdlg = $winMain::win->FileSelect(
      -title   => 'Save as',
      -create  => '1',
      -initialfile => $execFilename
   );
   my $fname = $sfdlg->Show();
   return if (($fname eq undef) || ($fname eq '*'));

   # Save everything
   ListingToProject($fname);
}

#----------
# OpenPrj()
#
# OpenProject

sub OpenPrj()
{
   # Open-dialog
   my $ofdlg = $winMain::win->FileSelect(
      -title  => 'Select Existing Project File',
      -filter => "*$config::ProjectExt"
   );
   my $fname = $ofdlg->Show();
   return if ($fname eq undef);

   # Read file
   _CleanUpWidgets();
   ProjectToListing($fname);
   _ProgressFinished($fname);
   $execFilename = undef;
}

#-------------------
# SearchResultGoto()
#
# Double-click on a search-result row

sub SearchResultGoto()
{
   my $srow = $SearchResult{listbox}->get('active');
   MoveRowSelection($srow);
}

#-------------
# SearchText()
#
# Searchs for a user-specified string

sub SearchText()
{
   # Dialog
   return if (($SearchText{dlg}->Show() eq $SearchText{btn}->[0]) ||
              ($SearchText{txt} eq ''));

   # Initialisation
   $SearchResult{top}->withdraw();
   $SearchResult{listbox}->delete(0,'end');
   my @sargs;
   push(@sargs,'-nocase') if (!$SearchText{case});
   push(@sargs,'-regexp') if ($SearchText{regexp});
   push(@sargs,'-forward');

   # Search
   my $curRow = 1;
   my $lastRow = 0;
   $SearchResult{matches} = 0;

   while (my $sIndex = $winMain::text->search(@sargs,$SearchText{txt},"$lastRow.0")) {
      ($curRow = $sIndex) =~ s/\..+//;
      last if ($curRow < $lastRow);
      $SearchResult{listbox}->insert('end',$curRow);
      $SearchResult{matches}++;
      $lastRow = $curRow+1;
   }

   # Handle results (incl. window)
   if ($SearchResult{matches} != 0) {
      # Active entry
      if ($SearchText{reswin} || ($SearchText{dir} eq 'down')) {
         $SearchResult{listbox}->activate(0);
      } else {
         $SearchResult{listbox}->activate('end');
      }
      # Unhide result-window or select row
      if ($SearchText{reswin}) {
         $SearchResult{top}->deiconify();
      } else {
         SearchResultGoto();
      }

   # Nothing found
   } else {
      ErrBox("Nothing found!");
   }
}

#------------------
# SearchResultWin()
#
# Shows the result-window

sub SearchResultWin()
{
   $SearchResult{top}->deiconify() if ($SearchText{reswin});
}

#-------------
# SearchNext()
#
# Jumps to the next search result-entry

sub SearchNext()
{
   # Nothing found - abort
   return if ($SearchResult{matches} == 0);

   # Get current item-no. and deselect it
   (my $actItem = $SearchResult{listbox}->index('active')) =~ s/\..+//;
   $SearchResult{listbox}->selection('clear','active');

   # dir: down
   if ($SearchText{dir} eq 'down') {
      $actItem++;
      $actItem = 0 if ($actItem == $SearchResult{matches});

   # dir: up
   } else {
      if ($actItem == 0) {
         $actItem = 'end';
      } else {
         $actItem--;
      }
   }

   # Activate and jump to it
   $SearchResult{listbox}->activate($actItem);
   $SearchResult{listbox}->selection('set','active');
   $SearchResult{listbox}->see('active');

   SearchResultGoto();
}

#--------------
# GotoCodeLoc()
#
# Dialog to enter a offset and jumps (if possible) to it

sub GotoCodeLoc()
{
   return if ($GotoCodeLoc{dlg}->Show() eq $GotoCodeLoc{btn}->[0]);

   # Is valid address ?
   my $startIndex = IsCodeAddr($GotoCodeLoc{ofs});
   if ($startIndex == 0) {
      ErrBox("$GotoCodeLoc{ofs} isn't a valid address");
      return;
   }

   # Search
   my $listAddr = '^:'.(($GotoCodeLoc{ofs} !~ /^0/) ? '0' : '').$GotoCodeLoc{ofs};
   my $addrIndex = $winMain::text->search(
      '-forward',
      '-nocase',
      '-regexp',
      $listAddr,
      $startIndex
   );
   if ($addrIndex ne undef) {
      MoveRowSelection($addrIndex);
   } else {
      ErrBox("$GotoCodeLoc{ofs} is a valid address, but this version of LDasm "
            ." can't interpolate between addresses");
   }
}

#-----------------
# GotoCodePoint($)
#
# Jumps to the code-start of entry-point

sub GotoCodePoint($)
{
   my $alias = shift;

   # Get row the section starts
   my %SectionAlias = (
      'start' => '.init',
      'ep'    => '.text'
   );
   return if (!exists $SectionAlias{$alias});
   my $secRow = GetSectionRow($SectionAlias{$alias});
   return if ($secRow eq undef);

   # Skip unnecessary rows
   $secRow =~ s/\..+//;
   while ($winMain::text->get("$secRow.0") ne ':') {
      $secRow++;
   }
   MoveRowSelection($secRow);
}

#-------------------
# JumpToReference($)
#
# Jumps to the reference if double-clicked
# Not in Export list, because both winFunc & winDlg link each other
#
# in:  toplevel-reference
# out: -

sub JumpToReference($)
{
   my $reftl = shift;

#   print "Type: $reftl->{type}\n";
   my $rstr = $reftl->{listbox}->get('active');
#   print "Active: '$rstr'\n";
   return if ($rstr eq undef);
   my $symref = $asmps::refsType{$reftl->{type}};
   return if ($symref eq undef);
   $symref = $symref->{$rstr};
   return if ($symref eq undef);

   if (ref $symref eq 'ARRAY') {
      if (exists $refDlgEntry{$symref}) {
         $refDlgEntry{$symref}++;
         $refDlgEntry{$symref} = 0 if ($refDlgEntry{$symref} == @{$symref});
      } else {
         $refDlgEntry{$symref} = 0;
      }
#      print "Array: ".@{$symref}." entries\n";
      # Move selection
#      print "Num: $refDlgEntry{$symref} - $symref->[$refDlgEntry{$symref}]\n";
      MoveRowSelection($symref->[$refDlgEntry{$symref}]);
#      print "> ".@{$symref}." Eintrge\n";
   } else {
#      print "NAA: $symref\n";
      MoveRowSelection($symref);
#      print "> 1 Eintrag\n";
   }
}

#--------------
# ReferenceCopy
#
# Copies the selected or all entries

sub ReferenceCopy
{
   my ($mode,$reftl) = @_;

   # Copy all
   if ($mode eq 'all') {
      for ($reftl->{listbox}->get(0,'end')) {
         print "$_\n";
      }

   # Selected
   } else {
      my $rcnt = $reftl->{listbox}->get('active');
      print "$rcnt\n" if ($rcnt ne undef);
   }
}

#--------------
# CopySelCode()
#
# Copies a single code-line

sub CopySelCode()
{
   if ($numRows > 0) {
      my $cline = $winMain::text->get("$winMain::currentRow.0","$winMain::currentRow.end");
      $cline =~ s/\s{2,}/ /g;
      print "$cline\n" if ($cline ne undef);
   }
}

#-------------
# ExecuteRefOp
#
# Executes a call or jump

sub ExecuteRefOp
{
   my $type = shift;

   # Not the required dereferencation-type
   return if (($refDest{'to'} eq undef) || ($refDest{'type'} ne $type));

   # Move selection and update deref-history
   push(@{$refDest{'history'}},{type => $type,row => $winMain::currentRow});
   MoveRowSelection($refDest{'to'});
}

#--------------
# ExecuteRefRet
#
# Returns from a jump or call

sub ExecuteRefRet
{
   my $type = shift;

   return if (($refDest{'history'} eq undef) ||
              (@{$refDest{'history'}} == 0)  ||
              ($refDest{'history'}->[-1]->{'type'} ne $type));

   my $th = pop @{$refDest{'history'}};
   MoveRowSelection($th->{'row'});
}



1;
